use crate::database;
use crate::usimp::*;
use std::collections::{HashMap, HashSet};
use std::ops::Deref;
use std::sync::Arc;
use tokio::sync::{mpsc, Mutex};
use tokio::sync::mpsc::Sender;

#[derive(Clone, Eq, Hash, PartialEq)]
struct Subscription {
    session: Uuid,
    req_nr: Option<u64>,
}

static mut SUBSCRIPTIONS: Option<Arc<Mutex<HashMap<Subscription, mpsc::Sender<Event>>>>> = None;
static mut ACCOUNTS: Option<Arc<Mutex<HashMap<Uuid, HashSet<Subscription>>>>> = None;

pub fn init() {
    unsafe {
        SUBSCRIPTIONS = Some(Arc::new(Mutex::new(HashMap::new())));
        ACCOUNTS = Some(Arc::new(Mutex::new(HashMap::new())));
    }
}

pub async fn subscribe_account(session: &Session, req_nr: Option<u64>) -> Result<mpsc::Receiver<Event>, Error> {
    let account = get_account(session)?;
    let sub = Subscription {session: session.id, req_nr};

    let (tx, rx) = mpsc::channel::<Event>(64);
    unsafe {
        let mut subs = SUBSCRIPTIONS.as_ref().unwrap().lock().await;
        let mut accs = ACCOUNTS.as_ref().unwrap().lock().await;
        match accs.get_mut(&account.id) {
            Some(set) => {
                set.insert(sub.clone());
            }
            None => {
                let mut set = HashSet::new();
                set.insert(sub.clone());
                accs.insert(account.id, set);
            }
        }
        match subs.get_mut(&sub) {
            Some(_) => return Err(Error::new(
                ErrorKind::SubscriptionError,
                ErrorClass::ClientError,
                None
            )),
            None => {
                subs.insert(sub.clone(), tx);
            }
        }
    }

    Ok(rx)
}

pub async fn unsubscribe_account(session: &Session, req_nr: Option<u64>) -> Result<(), Error> {
    let account = get_account(session)?;
    let sub = Subscription {session: session.id, req_nr};

    unsafe {
        let mut subs = SUBSCRIPTIONS.as_ref().unwrap().lock().await;
        let mut accs = ACCOUNTS.as_ref().unwrap().lock().await;
        match accs.get_mut(&account.id) {
            Some(set) => {
                set.remove(&sub);
            }
            None => {},
        }
        subs.remove(&sub);
    }

    Ok(())
}

pub async fn push_room(room_id: &Uuid, event: Event) -> Result<(), Error> {
    let backend = database::client().await?;
    let accounts = match backend {
        database::Client::Postgres(client) => {
            let res = client
                .query(
                    "SELECT account_id \
                    FROM member \
                    WHERE room_id = $1;",
                    &[&room_id],
                )
                .await?;
            let mut acc: Vec<Uuid> = Vec::new();
            for row in res {
                acc.push(row.get(0));
            }
            acc
        }
    };

    let mut room: Vec<mpsc::Sender<Event>> = Vec::new();

    for account in accounts {
        unsafe {
            let subs = SUBSCRIPTIONS.as_ref().unwrap().lock().await;
            let accs = ACCOUNTS.as_ref().unwrap().lock().await;
            if let Some(acc_subs) = accs.get(&account) {
                for sub in acc_subs {
                    match subs.get(sub) {
                        None => {}
                        Some(tx) => {
                            room.push(tx.clone());
                        }
                    }
                }
            }
        }
    }

    for tx in room {
        let _res = tx.send(event.clone()).await;
    }

    Ok(())
}
